/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/* Glob -- expand meta-characters in file names -- by Guido van Rossum

 * This is a filename 'globbing' routine that I wrote.  Besides expanding
 * *, ? and [...] exactly like sh(1), it also replaces initial ~ or ~user
 * by user's home directory, replaces initial $var by the contents of the
 * environment variable, and has a quoting mechanism whereby any of the
 * magic characters can be quoted by '\' (including '\' itself).

 * This software is copyright (c) 1986 by Stichting Mathematisch Centrum.
 * You may use, modify and copy it, provided this notice is present in all
 * modified or unmodified copies, and provided you don't make money for it.

 *** You may only distribute the set of files mentioned above as a whole. ***

 * If you find bugs, or have any other questions, please mail them to the
 * author, whose address is:

 Guido van Rossum
 CWI, dept. AA
 Kruislaan 413
 1089 SJ  Amsterdam

 Electronic mail: guido@mcvax.UUCP
 ***************************************************************************/

/*
 * 'exParsePath' routine -- match *, [] and ? in filename.
 * Extra services: initial ~username is replaced by username's home dir,
 * initial ~ is replaced by $HOME, $var is replaced by
 * return value of getenv("var"), {,,,} is expanded.
 * Quoting convention: \ followed by a magic character inhibits its
 * special meaning.
 */

#include <stdio.h>
#include <string.h>
#include <pwd.h> 
#include <ctype.h>
#include <sys/types.h> 

#include <strings.h>
#include <sys/dir.h> 

#include <exBasics.h> 
#include <exErrors.h>

struct passwd *getpwnam();

static char *expand_home();
static char *ParseVariables();
static char *exSubstHome();
static char *exSubstVar();
static int exGlob();
static ExpandList();

#define SEP '/'

#define MAXPATH 1024
#define MAXBASE 256
char *mlist[256];

/**************************************************************************
 *                      Public Functions
 ***************************************************************************/

/* exExpand ******************************************************************
 * interprets '{}'in pathname, then calls 'exGlob' for matching *, [], ?;
 * returns number of matches which are stored in 'Buf' separated by blanks.
 * If no matches were found, the pattern is copied to the buffer after \ 
 * quoting. 
 ****************************************************************************/ 

exExpand(WorkDir, Pat, Buf, sizeofbuf)
     char *WorkDir;           /* current working directory */
     char *Pat;               /* pattern to be expanded */
     char *Buf;               /* holds resulting matches separated by blanks */
     int sizeofbuf;           /* maximum size of buffer */
{
  char *expList[128], *CompList[256];       
  int nu = 0, n, l;
  register i, j, num;

  if ((l = ExpandList(Pat, expList)) == -1) {
    strcat(Buf, Pat);
    return(0);
  }

  for (i = 0; i < 256; i++)
    CompList[i] = (char *) NULL;

  for (i = 0; i < l; i++) {
    if (strpbrk(expList[i],"*?\\[]") != NULL &&
	(n = exGlob(WorkDir, expList[i], CompList)) != 0) {
      
      free(expList[i]);
      
      for (j = 0, num = 0; j < 256 && num < n; j++) {
	
	if (strlen(Buf) < sizeofbuf) {
	  if (CompList[j]) {
	    strcat(Buf, CompList[j]);
	    strcat(Buf, " ");            /* matches are separated by blanks */
	    free(CompList[j]);
	    num++;
	  }
	} else {
	  setError(ERR_MAX);
	  return(-1);
	}
      }
      
      nu += n;
    } else {
      strcat(Buf, expList[i]);
      strcat(Buf, " ");
      free(expList[i]);
    }  
  }
  
  return(nu);
}

/* exParsePath ***************************************************************
 * match *, [] and ? in filename;
 * Extra services: initial ~username is replaced by username's home dir,
 * initial ~ is replaced by $HOME, $var is replaced by return value of 
 * getenv, {,,,} is expanded.
 * Quoting convention: \ followed by a magic character inhibits its
 * special meaning.
 ****************************************************************************/

exParsePath(path, prefix, textBuf, textBufSize)
     char *path;         /* pattern to be parsed */
     char **prefix;	 /* path prefix in case of '~' and var. expansion */
     char *textBuf;      /* holds resulting matches separated by blanks */
     int textBufSize;    /* maximum size of buffer */
{
  register cnt;
  char *ip, *p, *ptr;
  char *getenv();

  if (! (ip = expand_home(path)))	 	/* expand '~' */
    return (-1);
  
  p = ip;

  if (! (ip = ParseVariables(p, getenv))) {	/* expand vars. (${..}) */
    free(p);
    return (-1);
  }

  if (! (ptr = strrchr(ip, '/'))) {		/* expand in current dir. */
    cnt = exExpand(".", ip, textBuf, textBufSize);
    *prefix = NULL;
  } else {					/* this is the 'normal' case */
    *ptr++ = EOS;				/* strip off last component */
    if (*ptr) {					/* something to expand */
      cnt = exExpand(ip, ptr, textBuf, textBufSize);
      *prefix = strCopy(ip);			/* remember prefix of exp. */
    } else {					/* nothing more to expand */
      if (strcmp(ip, path) == 0) {		/* path has not changed */
	cnt = 0;
      } else {					/* ~ or vars expanded */
	cnt = 1;				/* say that we've expanded */
	strncpy(textBuf, ip, textBufSize);	/* return expanded path */
      }
      *prefix = NULL;				/* no new prefix */
    }
  }
  free(ip);
  free(p);
  return(cnt);
}

/* convertRegExp *************************************************************
 * we have to change the regular expression as the syntax of the regular 
 * expressions that the shell uses is a bit different from those understood 
 * by the general regular expression evaluation routines. The differences are 
 * described below. They consist merely of the different interpretation of 
 * certain special characters.
 *
 * SHELL			REGEX
 * *                            .*	any number of occurances of any char 
 * .				\.      dot has a special meaning in regex
 * ?				.       match exactly one character 
 * {pat1,pat2,...patn}                  match any string in list. Must be 
 *   					simulated by doing a match with each 
 *   			                of the patterns.
 *****************************************************************************/

convertRegExp(rexp, shellRexp)
     char *rexp;	  	/* reg. expr. to be converted */
     char *shellRexp;		/* holds the converted reg. expr. */
{
  register char *sp = rexp;	/* pointer in the in-expression */
  register char *ssp;		/* pointer in the out-expression */
  int escaped = 0;		/* flag that indicates whether the next char 
				   should be taken literally */
  ssp = shellRexp;
  *ssp++ = '^';			/* match always from beginning */
  while (*sp) {
    switch (*sp) {
    case '*': 			/* '*' --> .* */
      if (escaped) {
	escaped = 0;
	*ssp++ = '\\';
      } else
	*ssp++ = '.';
      *ssp++ = '*';
      break;
    case '.':			/* . --> \. */
      *ssp++ = '\\';
      *ssp++ = '.';
      break;
    case '?':
      if (escaped) {
	*ssp++ = '?';
	escaped = 0;
      } else 
	*ssp++ = '.';
      break;
    case '\\':			/* toggle escape-flag */
      if (escaped) {
	*ssp++ = '\\';
	escaped = 0;
      } else
	escaped = 1;
      break;
    default:			/* copy normal chars */
      *ssp++ = *sp;
      if (escaped)
	escaped = 0;
      break;
    }
    sp++;
  }
  *ssp = EOS;
}

/**************************************************************************
 *                      Private Functions
 ***************************************************************************/

/* expand_home ***************************************************************
 * expand ~[user] in file name (HOME directory)
 ****************************************************************************/

static char *expand_home(n)
     char *n;
{
  char tmp[MAXPATH], Home[MAXBASE];
  register char *str;

  if (*n == '~') { 
    str = exSubstHome(n, Home);
    if (!Home || ! *Home)          	/* HOME or user name not defined */
      return ((char *) 0);
    else {
      strcpy(tmp, Home);
      strcat(tmp, str);
      return(strCopy(tmp));
    }
  } else
    return(strCopy(n));
}

/* ParseVariables ***********************************************************
 * does all '$' substitutions in path and returns generated path
 ****************************************************************************/

static char *ParseVariables(pattern, getVar)
     char *pattern;         /* pattern containing variables */
     char *(*getVar)();     /* pointer to function returning variable value */
{
  char *point, *start, 
       *Var, *Prefix, *Rest;

  start = point = pattern;
  Prefix = (char *) malloc(MAXPATH);
  *Prefix = EOS;

  while (*pattern != EOS) 

    if (*pattern == '$') {

      Rest = makeStr(point, pattern, MAXBASE);
      strcat(Prefix, Rest);
      free(Rest);

      Var = (char *) malloc(MAXBASE);
      if ((pattern = exSubstVar(pattern, Var, getVar)) == NULL) {
	free(Var);
	return(NULL);
      }

      strcat(Prefix, Var);
      free(Var);
      point = pattern;
    }
    else
      pattern++; 

  if (*Prefix != EOS) {
    Rest = makeStr(point, pattern, MAXBASE);
    strcat(Prefix, Rest);
    free(Rest);
  } else 
    strcpy(Prefix, start);

  return(Prefix);
}

/* exSubstHome ************************************************************    
 * does  '~' and '~[user]' substitution
 ***************************************************************************/

static char *exSubstHome(path, buf)
  char *path;                            /* path beginning with '~' */
  char *buf;                             /* will contain result of expansion */
{
  char *p, *User;
  struct passwd *pwn;
  char *cuserid();
  
  p = ++path;
  while (*p != EOS && *p != SEP)
    ++p;

  if (p == path)                                       /* expand '~/' */
    User = cuserid((char *) 0);
  else                                                 /* expand '~[user]/' */
    User = makeStr(path, p, 100);

  if ((User) && (pwn = getpwnam(User))) {
    strcpy(buf, pwn->pw_dir);
    if (buf[strlen(buf) - 1] == '/')
      buf[strlen(buf) - 1] = EOS;
    path = p;
    free(User);
  } else {
    setError(ERR_NOEXIST);
    --path; 
  }

  return (path);                   /* points to first '/' after ~-expression */
}

/* exSubstVar *************************************************************    
 * does  '$' substitution
 **************************************************************************/

static char *exSubstVar(path, buf, varFunc)
     char *path;           /* path beginning with '$' */
     char *buf;            /* will contain '$' or result of substitution */
     char *(*varFunc)();   /* pointer to function returning variable value */  
{
  char *p, *Name, *var;

  p = ++path;                           /* position after '$' */

  if ((p = strpbrk(p, "/,}\0")) == NULL) {
                           /* end of variable name is indicated by '/,}\0' */
    p = path;
    while (*p != EOS)
      ++p;
  }

  Name = makeStr(path, p, 100);

  if (Name != NULL) {
    if ((var = varFunc/*getenv*/(Name)) != NULL)  
      strcpy(buf, var);
    else {
      setError(ERR_NOEXIST);
      return(NULL);
    }
    path = p;            /* position behind recognized end of variable name */
    free(Name);
  }
  else                                            /* no name behind '$' */
    *buf = '$';

  return(path);
}

/* matchFiles ****************************************************************
 * evaluates regular expression and returns number and list of matches
 ****************************************************************************/

static matchFiles(path, rexp, m_list, matches)
     char *path;                         /* directory in which reg.exp. will
					    be evaluated */
     char *rexp;                         /* regular expression to compile */
     char *m_list[];                      /* holds matches found */
     int *matches;                       /* holds number of matches */
{
  DIR *dirp;
  struct direct *dp;
  char *re_comp();
  char shellRexp[1024];			 /* holds the converted reg. exp. */
  int cnt = 0;

  convertRegExp(rexp, shellRexp);	 /* convert the reg.exp. syntax */

  if (re_comp(shellRexp)) {		 /* compile the regular expression */
    *matches = -1;
    return(setError(ERR_ABORT));
  }

  if ((dirp = opendir(path)) == (DIR *) 0) {  
    *matches = -1;
    return(setError(ERR_OPEN));
  }

  while ((dp = readdir(dirp)) != (struct direct *) 0)
    if (re_exec(dp->d_name)) 		/* do the match */
      m_list[cnt++] = strCopy(dp->d_name);

  closedir(dirp);
  *matches = cnt;
  return(setError(ERR_NOERR));
}

/* exGlob ******************************************************************
 *  The main 'exGlob' routine: calls 'matchFiles' (relative to working
 *  directory if necessary) to do the pattern matching and quoting.            
 *  Returns 0 if file not found, number of matches otherwise.          
 *  The matches found are stored in an array of pointers to char.            
 ****************************************************************************/

static int exGlob(workd, pat, complist)
     char *workd;          /* current working directory */
     char *pat;            /* pattern containing regular expressions */
     char *complist[];     /* bearbeitete und noch auszuwertende Pfade */
{
  register cnt = 0,        /* letzter belegter Index in complist */
           count = 0, i, j = 0, matches;
  int num;
  char *Prefix,            /* working directory if path starts with '/' */
       *rest,              /* regexp auf der aktuellen Ebene */
       *Mlist[MAXPATH],    /* Liste fuer gefundene matches */
       *p, 
       FullName[MAXPATH],  /* vollstaendiger Pfadname */
       *Makestr;           /* fuer Ergebnis von 'makestr' */
       char *start;     /* Pointer auf Anfang des aktuell untersuchten Pfads */


  complist[0] = (char *) malloc(MAXBASE);
  matches = 0;
  Prefix = NULL;

  if (*pat != '/') { 
    Prefix = strCopy(workd);
    *complist[0] = EOS;
  }
  else
    *complist[0] = SEP;

  if (*pat == SEP)
    pat++;
  p = pat;              /* pat steht auf Anfang des zu untersuchenden Pfades */
  count = ++cnt;
  matches++;

  while (*pat != EOS) {

    while (*pat != SEP && *pat != EOS) 
       pat++;

    rest = makeStr(p, pat, 100);
    strcat(rest, "$");

    if (*pat == SEP)
      pat++;
    p = pat;                                  /* Anfang einer neuen Ebene */

    for (i = 0; i < count; i++) {

      start = complist[i];

      if (complist[i]) {
	while (*complist[i] != EOS)
	  complist[i]++;

	Makestr = makeStr(start, complist[i], MAXBASE);
	if (Prefix) {
	  makeFullName(Prefix, Makestr, FullName, MAXBASE);	  
	  matchFiles(FullName, rest, Mlist, &num);
	}
	else
	  matchFiles(Makestr, rest, Mlist, &num);

	if (num > 0) {
	  for (j = 1; j < num; j++) {

	    if (*Makestr != EOS) {
	      makeFullName(Makestr, Mlist[j], FullName, MAXBASE);
	      complist[cnt++] = strCopy(FullName);
	    }
	    else
	      complist[cnt++] = strCopy(Mlist[j]);
	   
	    matches++;
	  }

	  free(complist[i]);

	  if (*Makestr != EOS) {
	    makeFullName(Makestr, Mlist[0], FullName, MAXBASE);
	    complist[i] = strCopy(FullName);
	  }
	  else
	    complist[i] = strCopy(Mlist[0]);
	}
        else {                           /* no match oder Fehler, loeschen */
	  free(complist[i]);
	  complist[i] = NULL;
	  matches--;
	}	      
      }
    }

    count = cnt;
    free(rest);
    free(Makestr);      
  }

  free(Prefix);
  return matches;
}

/* ExpandList ***************************************************************
 * interprets '{,,,}' and returns number of generated paths
 ****************************************************************************/

static ExpandList(pattern, pathlist)
     char *pattern;
     char *pathlist[];
{
  char *start, *openpar, *element, *contents, *Prefix, *p, *rest;
  int i = 0, cnt = 0, n = 1;
  Boolean subst = False;

  pathlist[i] = strCopy(pattern);

  while (cnt < n || subst == True) {
    cnt = n;
    for(i = 0; i < cnt; i++) {

      start = pathlist[i];

      if (((p = rindex(pathlist[i], '{')) != NULL) 
	       && *(p - 1) != '$') {
	openpar = p;      
	pathlist[i] = openpar;       /* pathlist auf rechtester { ? */

	while (*pathlist[i]++ != '}') {      /* Search next closing brace */
	  if (*pathlist[i] == EOS) {
	    setError(ERR_VALUE);
	    /* printf("Missing }\n"); */
	    return(-1);
	  }
	}
	
	rest = strCopy(pathlist[i]);
	contents = makeStr(openpar + 1, pathlist[i], MAXBASE);
    
	if ((element = strtok(contents, ",}")) != NULL) {

	  Prefix = makeStr(start, openpar, MAXBASE);
	  free(pathlist[i]);
	  pathlist[i] = (char *) malloc(MAXPATH);

	  if ((strlen(Prefix) + strlen(element) + strlen(rest)) > MAXBASE) {
	    setError(ERR_MAX);
	    return(-1);
	  }

	  strcpy(pathlist[i], Prefix);
	  strcat(pathlist[i], element);
	  strcat(pathlist[i], rest);

	  element = NULL;
	  subst = True;

	  while ((element = strtok(element, ",}")) != NULL) {
	    pathlist[n] = (char *) malloc(MAXPATH);

	    if ((strlen(Prefix) + strlen(element) + strlen(rest)) > MAXBASE) {
	      setError(ERR_MAX);
	      return(-1);
	    }

	    strcpy(pathlist[n], Prefix);
	    strcat(pathlist[n], element);
	    strcat(pathlist[n++], rest);
	 
	    element = NULL;
	  }
	  
	  free(Prefix);
	}

	free(rest);
	free(contents);
      }
      else
	subst = False;
    }
  }

  return(n);
}




