/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/* X & Xm headers */
#include <X11/IntrinsicP.h>
#include <Xm/BulletinB.h>
#include <X11/Shell.h>

#include <exBasics.h>
#include <exLayout.h>
#include "exThermometer.h"
#include <strings.h>
#include <string.h>

static void therm_redraw();
static void draw_bar();
static void therm_init();
static void therm_destroy();
static void redraw_therm();

static GC draw_gc = NULL;
static GC fill_gc = NULL;

static exThermometer *thermometer = (exThermometer *) NULL;
 
/* initThermometer: *********************************************************
 * initialize a new thermometer structure; if the first time, the 
 * associated widgets are created; otherwise they are adapted by providing  
 * new parameters
 ***************************************************************************/

void initThermometer(d, parent, x, y, c, msg, font)
     Display *d;
     Widget parent;
     int x, y;
     int c; 
     char *msg;
     XFontStruct *font;
{
  static Boolean first = True;
  int lines, w, i, cnt;
  char *fmsg;
  Arg args[10];
  int fh; 
  XmFontList fl = NULL;
  register char *ptr;

  static XtTranslations ptt = NULL;

  static char thermometer_translations[] = 
   "<Expose>		: therm_redraw()";

  static XtActionsRec thermometer_actions[] = {
    { "therm_redraw",    therm_redraw}
  };

  if (! ptt) {
    XtAddActions(thermometer_actions, 1);
    ptt = XtParseTranslationTable(thermometer_translations);
  }

  if (! draw_gc)
    therm_init(d);
  
  if (! thermometer)
    thermometer = (exThermometer *) Malloc(sizeof(exThermometer));

  lines = fformatText(msg, c, &fmsg);
  
  thermometer->filled_w = 0;
  thermometer->d = d;

  if (! first) {				/* free old stuff */
    for (i = 0; i < thermometer->lines; i++)
      free(thermometer->txt[i]);
    free(thermometer->txt);
  }

  thermometer->txt = (char **) calloc(lines, sizeof(char *));

  if (first) {
    thermometer->t_s = XtCreatePopupShell("thermometer", 
		     overrideShellWidgetClass, parent, NULL, 0);
    
    cnt = 0;
    XtSetArg(args[cnt], XmNuserData, thermometer); cnt++;
    thermometer->t_w = XtCreateManagedWidget("thermometer", 
		     xmBulletinBoardWidgetClass, thermometer->t_s, args, cnt);

    XtOverrideTranslations(thermometer->t_w, ptt);

    XtAddCallback(thermometer->t_w, XmNdestroyCallback, 
		                       therm_destroy, thermometer);

    if (! font) {      			/* get font resource */
      cnt = 0;
      XtSetArg(args[cnt], XmNtextFontList, &fl); cnt++;
      XtGetValues(thermometer->t_w, args, cnt);
    
      if (! fl) 
	thermometer->font = exBoldFont(d);
      else 
	thermometer->font = getFont(fl);
    
      if (! thermometer->font) 
	fatal("initThermometer", "Cannot load font !", NULL);
    }
    first = False;
  }

  for (ptr = strtok(fmsg, "\n"), i = 0, w = 0; ptr; ptr = strtok(NULL, "\n")) {
    thermometer->txt[i++] = strCopy(ptr);
    w = MAX(w, exTextWidth(thermometer->font, ptr));
  }

  fh = exFontHeight(thermometer->font);

  thermometer->lines = lines;
  thermometer->bar_h = 3 * fh / 2;
  thermometer->bar_w = w;
  thermometer->bar_x = TGAP;
  thermometer->bar_y = lines * fh + 2 * TGAP;

  cnt = 0;
  XtSetArg(args[cnt], XmNx, x); cnt++;
  XtSetArg(args[cnt], XmNy, y); cnt++;
  XtSetArg(args[cnt], XmNwidth, w + 2 * TGAP); cnt++;
  XtSetArg(args[cnt], XmNheight, 
	   3 * TGAP + lines * fh + thermometer->bar_h); cnt++;

  XtSetValues(thermometer->t_s, args, cnt);
}

/* setThermometer: **********************************************************
 * is called whenever the indicated progress should be adjusted. The integer 
 * parameter 'progress' should be between 0 and 100 indicating the percentage 
 * of progress. The thermometer is moved to the new location indicated by 
 * x an y
 ****************************************************************************/

void setThermometer(x, y, progress)
     int x, y, progress;
{
  if (progress > 100)
    progress = 100;
  else if (progress < 0)
    progress = 0;

  if (x >= 0 && y >= 0)
    XtMoveWidget(thermometer->t_s, x, y);

  if (! (thermometer->flags & TH_MAPPED)) {	/* map window if required */
    XtPopup(thermometer->t_s, XtGrabNone);
    thermometer->flags |= TH_MAPPED;
    exWait(5);
  }

  thermometer->filled_w = thermometer->bar_w * progress / 100;
  
  redraw_therm(thermometer);
  XSync(thermometer->d, False);
}

/* hideThermometer: *********************************************************
 * hides the thermometer from the screen; waits ticks * 50 ms before hiding 
 ***************************************************************************/

void hideThermometer(delay)
     int delay;
{
  exWait(delay);
  XtPopdown(thermometer->t_s);
  thermometer->flags &= ~TH_MAPPED;			/* mark as unmapped */
  XSync(thermometer->d, False);
}

/*****************************************************************************
 *
 *  		internally used functions 
 *
 *****************************************************************************/

/* therm_redraw: **************************************************************
 * this function is called whenever an exposure event happened with the 
 * thermometer window. It reddaws the thermometer 
 *****************************************************************************/
/*ARGSUSED*/
static void therm_redraw(w, event, params, num_params)
     Widget w;
     XExposeEvent *event;
     String *params;
     Cardinal *num_params;
{
  exThermometer *therm;
  Arg args[1];
  int cnt;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &therm); cnt++;
  XtGetValues(w, args, cnt);

  redraw_therm(therm);
}

static void redraw_therm(therm)
     exThermometer *therm;
{
  int i, y;
  Window w;
  int fh; 
   
  w = XtWindow(therm->t_w);

  if (! (therm->flags & TH_MAPPED))			/* window not mapped */
    return;						/* do not map it */
  
  fh = exFontHeight(therm->font);
  y = exFontHeight(therm->font) + TGAP;
  XSetFont(therm->d, draw_gc, therm->font->fid);

  XtPopup(therm->t_s, XtGrabNone);

  for (i = 0; i < therm->lines; i++) {
    XDrawString(therm->d, w, draw_gc, TGAP, y, therm->txt[i], 
						strlen(therm->txt[i]));
    y += fh;
  }

  XDrawRectangle(therm->d, w, draw_gc, 
		 therm->bar_x, therm->bar_y, therm->bar_w, therm->bar_h);

  draw_bar(therm);
}

/* draw_bar: ****************************************************************
 * draws the bar within a thermometer
 ***************************************************************************/

static void draw_bar(therm)
     exThermometer *therm;
{
  Window w;
  
  w = XtWindow(therm->t_w);

  /* draw filled area */
  if (therm->filled_w > 2)
    XFillRectangle(therm->d, w, fill_gc, therm->bar_x + 1, therm->bar_y + 1,
		   therm->filled_w - 2, therm->bar_h - 2);
  
  /* clear tail */
  XClearArea(therm->d, w, therm->bar_x + 1 + therm->filled_w, therm->bar_y + 1,
	     therm->bar_w - therm->filled_w - 2, therm->bar_h - 2, False);
}

/* therm_destroy: ***********************************************************
 * remove the thermometer from the screen and free the data-structures. 
 ****************************************************************************/
/*ARGSUSED*/
static void therm_destroy(wd, client_data, call_data)
     Widget wd;
     caddr_t client_data;
     caddr_t call_data;
{
  register i;
  
  for (i = 0; i < thermometer->lines; i++)
    free(thermometer->txt[i]);
  free(thermometer->txt);
  free(thermometer);
  thermometer = (exThermometer *) NULL;
}
  
/* therm_init: **************************************************************
 * intialize the gc's used during thermometer management 
 ***************************************************************************/ 

static void therm_init(d)
     Display *d;
{
  draw_gc = XCreateGC(d, XDefaultRootWindow(d), 0, 0);
  XSetForeground(d, draw_gc, exBlackPixel(d));
  XSetBackground(d, draw_gc, exWhitePixel(d));
  XSetLineAttributes(d, draw_gc, 2, LineSolid, CapButt, JoinMiter);
  XSetFillStyle(d, draw_gc, FillSolid);
  XSetFunction(d, draw_gc, GXcopy);

  fill_gc = XCreateGC(d, XDefaultRootWindow(d), 0, 0);
  XSetForeground(d, fill_gc, exBlackPixel(d));
  XSetBackground(d, fill_gc, exWhitePixel(d));
  XSetFillStyle(d, fill_gc, FillTiled);
  XSetTile(d, fill_gc, exDark_greyPM(d));
  XSetFunction(d, fill_gc, GXcopy);
}
